@@ -120,7 +120,9 @@ radv_surface_has_scanout(struct radv_device *device, const struct radv_image_cre
    const struct radv_physical_device *pdev = radv_device_physical(device);
 
    if (info->bo_metadata) {
-      if (pdev->info.gfx_level >= GFX9)
+      if (pdev->info.gfx_level >= GFX12) {
+         return info->bo_metadata->u.gfx12.scanout;
+      } else if (pdev->info.gfx_level >= GFX9)
          return info->bo_metadata->u.gfx9.scanout;
       else
          return info->bo_metadata->u.legacy.scanout;
@@ -472,7 +474,13 @@ radv_patch_surface_from_metadata(struct radv_device *device, struct radeon_surf
 
    surface->flags = RADEON_SURF_CLR(surface->flags, MODE);
 
-   if (pdev->info.gfx_level >= GFX9) {
+   if (pdev->info.gfx_level >= GFX12) {
+      surface->u.gfx9.swizzle_mode = md->u.gfx12.swizzle_mode;
+      surface->u.gfx9.color.dcc.max_compressed_block_size = md->u.gfx12.dcc_max_compressed_block;
+      surface->u.gfx9.color.dcc_data_format = md->u.gfx12.dcc_data_format;
+      surface->u.gfx9.color.dcc_number_type = md->u.gfx12.dcc_number_type;
+      surface->u.gfx9.color.dcc_write_compress_disable = md->u.gfx12.dcc_write_compress_disable;
+   } else if (pdev->info.gfx_level >= GFX9) {
       if (md->u.gfx9.swizzle_mode > 0)
          surface->flags |= RADEON_SURF_SET(RADEON_SURF_MODE_2D, MODE);
       else
@@ -748,20 +756,55 @@ radv_compose_swizzle(const struct util_format_description *desc, const VkCompone
    }
 }
 
-static void
-radv_query_opaque_metadata(struct radv_device *device, struct radv_image *image, unsigned plane_id,
-                           struct radeon_bo_metadata *md)
+void
+radv_image_bo_set_metadata(struct radv_device *device, struct radv_image *image, struct radeon_winsys_bo *bo)
 {
    const struct radv_physical_device *pdev = radv_device_physical(device);
    const struct radv_instance *instance = radv_physical_device_instance(pdev);
    static const VkComponentMapping fixedmapping;
+   const uint32_t plane_id = 0; /* Always plane 0 to follow RadeonSI. */
    const VkFormat plane_format = radv_image_get_plane_format(pdev, image, plane_id);
    const unsigned plane_width = vk_format_get_plane_width(image->vk.format, plane_id, image->vk.extent.width);
    const unsigned plane_height = vk_format_get_plane_height(image->vk.format, plane_id, image->vk.extent.height);
    struct radeon_surf *surface = &image->planes[plane_id].surface;
    const struct legacy_surf_level *base_level_info = pdev->info.gfx_level <= GFX8 ? &surface->u.legacy.level[0] : NULL;
+   struct radeon_bo_metadata md;
    uint32_t desc[8];
 
+   memset(&md, 0, sizeof(md));
+
+   if (pdev->info.gfx_level >= GFX12) {
+      md.u.gfx12.swizzle_mode = surface->u.gfx9.swizzle_mode;
+      md.u.gfx12.dcc_max_compressed_block = surface->u.gfx9.color.dcc.max_compressed_block_size;
+      md.u.gfx12.dcc_number_type = surface->u.gfx9.color.dcc_number_type;
+      md.u.gfx12.dcc_data_format = surface->u.gfx9.color.dcc_data_format;
+      md.u.gfx12.dcc_write_compress_disable = surface->u.gfx9.color.dcc_write_compress_disable;
+      md.u.gfx12.scanout = (surface->flags & RADEON_SURF_SCANOUT) != 0;
+   } else if (pdev->info.gfx_level >= GFX9) {
+      uint64_t dcc_offset =
+         image->bindings[0].offset + (surface->display_dcc_offset ? surface->display_dcc_offset : surface->meta_offset);
+      md.u.gfx9.swizzle_mode = surface->u.gfx9.swizzle_mode;
+      md.u.gfx9.dcc_offset_256b = dcc_offset >> 8;
+      md.u.gfx9.dcc_pitch_max = surface->u.gfx9.color.display_dcc_pitch_max;
+      md.u.gfx9.dcc_independent_64b_blocks = surface->u.gfx9.color.dcc.independent_64B_blocks;
+      md.u.gfx9.dcc_independent_128b_blocks = surface->u.gfx9.color.dcc.independent_128B_blocks;
+      md.u.gfx9.dcc_max_compressed_block_size = surface->u.gfx9.color.dcc.max_compressed_block_size;
+      md.u.gfx9.scanout = (surface->flags & RADEON_SURF_SCANOUT) != 0;
+   } else {
+      md.u.legacy.microtile =
+         surface->u.legacy.level[0].mode >= RADEON_SURF_MODE_1D ? RADEON_LAYOUT_TILED : RADEON_LAYOUT_LINEAR;
+      md.u.legacy.macrotile =
+         surface->u.legacy.level[0].mode >= RADEON_SURF_MODE_2D ? RADEON_LAYOUT_TILED : RADEON_LAYOUT_LINEAR;
+      md.u.legacy.pipe_config = surface->u.legacy.pipe_config;
+      md.u.legacy.bankw = surface->u.legacy.bankw;
+      md.u.legacy.bankh = surface->u.legacy.bankh;
+      md.u.legacy.tile_split = surface->u.legacy.tile_split;
+      md.u.legacy.mtilea = surface->u.legacy.mtilea;
+      md.u.legacy.num_banks = surface->u.legacy.num_banks;
+      md.u.legacy.stride = surface->u.legacy.level[0].nblk_x * surface->bpe;
+      md.u.legacy.scanout = (surface->flags & RADEON_SURF_SCANOUT) != 0;
+   }
+
    radv_make_texture_descriptor(device, image, false, (VkImageViewType)image->vk.image_type, plane_format,
                                 &fixedmapping, 0, image->vk.mip_levels - 1, 0, image->vk.array_layers - 1, plane_width,
                                 plane_height, image->vk.extent.depth, 0.0f, desc, NULL, NULL, NULL);
@@ -769,46 +812,10 @@ radv_query_opaque_metadata(struct radv_device *device, struct radv_image *image,
    radv_set_mutable_tex_desc_fields(device, image, base_level_info, plane_id, 0, 0, surface->blk_w, false, false, false,
                                     false, desc, NULL, 0);
 
-   ac_surface_compute_umd_metadata(&pdev->info, surface, image->vk.mip_levels, desc, &md->size_metadata, md->metadata,
+   ac_surface_compute_umd_metadata(&pdev->info, surface, image->vk.mip_levels, desc, &md.size_metadata, md.metadata,
                                    instance->debug_flags & RADV_DEBUG_EXTRA_MD);
-}
 
-void
-radv_init_metadata(struct radv_device *device, struct radv_image *image, struct radeon_bo_metadata *metadata)
-{
-   const struct radv_physical_device *pdev = radv_device_physical(device);
-
-   /* use plane 0, even when there are multiple planes, to follow radeonsi */
-   const unsigned plane_id = 0;
-   struct radeon_surf *surface = &image->planes[plane_id].surface;
-
-   memset(metadata, 0, sizeof(*metadata));
-
-   if (pdev->info.gfx_level >= GFX9) {
-      uint64_t dcc_offset =
-         image->bindings[0].offset + (surface->display_dcc_offset ? surface->display_dcc_offset : surface->meta_offset);
-      metadata->u.gfx9.swizzle_mode = surface->u.gfx9.swizzle_mode;
-      metadata->u.gfx9.dcc_offset_256b = dcc_offset >> 8;
-      metadata->u.gfx9.dcc_pitch_max = surface->u.gfx9.color.display_dcc_pitch_max;
-      metadata->u.gfx9.dcc_independent_64b_blocks = surface->u.gfx9.color.dcc.independent_64B_blocks;
-      metadata->u.gfx9.dcc_independent_128b_blocks = surface->u.gfx9.color.dcc.independent_128B_blocks;
-      metadata->u.gfx9.dcc_max_compressed_block_size = surface->u.gfx9.color.dcc.max_compressed_block_size;
-      metadata->u.gfx9.scanout = (surface->flags & RADEON_SURF_SCANOUT) != 0;
-   } else {
-      metadata->u.legacy.microtile =
-         surface->u.legacy.level[0].mode >= RADEON_SURF_MODE_1D ? RADEON_LAYOUT_TILED : RADEON_LAYOUT_LINEAR;
-      metadata->u.legacy.macrotile =
-         surface->u.legacy.level[0].mode >= RADEON_SURF_MODE_2D ? RADEON_LAYOUT_TILED : RADEON_LAYOUT_LINEAR;
-      metadata->u.legacy.pipe_config = surface->u.legacy.pipe_config;
-      metadata->u.legacy.bankw = surface->u.legacy.bankw;
-      metadata->u.legacy.bankh = surface->u.legacy.bankh;
-      metadata->u.legacy.tile_split = surface->u.legacy.tile_split;
-      metadata->u.legacy.mtilea = surface->u.legacy.mtilea;
-      metadata->u.legacy.num_banks = surface->u.legacy.num_banks;
-      metadata->u.legacy.stride = surface->u.legacy.level[0].nblk_x * surface->bpe;
-      metadata->u.legacy.scanout = (surface->flags & RADEON_SURF_SCANOUT) != 0;
-   }
-   radv_query_opaque_metadata(device, image, plane_id, metadata);
+   device->ws->buffer_set_metadata(device->ws, bo, &md);
 }
 
 void
@@ -875,75 +882,36 @@ radv_image_alloc_values(const struct radv_device *device, struct radv_image *ima
  * which requires to invalidate L2.
  */
 static bool
-radv_image_is_pipe_misaligned(const struct radv_device *device, const struct radv_image *image)
+radv_image_is_pipe_misaligned(const struct radv_image *image, const VkImageSubresourceRange *range)
 {
-   const struct radv_physical_device *pdev = radv_device_physical(device);
-   const struct radeon_info *gpu_info = &pdev->info;
-   int log2_samples = util_logbase2(image->vk.samples);
-
-   assert(gpu_info->gfx_level >= GFX10);
-
    for (unsigned i = 0; i < image->plane_count; ++i) {
-      VkFormat fmt = radv_image_get_plane_format(pdev, image, i);
-      int log2_bpp = util_logbase2(vk_format_get_blocksize(fmt));
-      int log2_bpp_and_samples;
+      const uint32_t first_mip_pipe_misaligned = image->planes[i].first_mip_pipe_misaligned;
 
-      if (gpu_info->gfx_level >= GFX10_3) {
-         log2_bpp_and_samples = log2_bpp + log2_samples;
-      } else {
-         if (vk_format_has_depth(image->vk.format) && image->vk.array_layers >= 8) {
-            log2_bpp = 2;
-         }
-
-         log2_bpp_and_samples = MIN2(6, log2_bpp + log2_samples);
-      }
-
-      int num_pipes = G_0098F8_NUM_PIPES(gpu_info->gb_addr_config);
-      int overlap = MAX2(0, log2_bpp_and_samples + num_pipes - 8);
-
-      if (vk_format_has_depth(image->vk.format)) {
-         if (radv_image_is_tc_compat_htile(image) && overlap) {
+      if (range) {
+         if (range->baseMipLevel + range->levelCount - 1 >= first_mip_pipe_misaligned)
             return true;
-         }
       } else {
-         int max_compressed_frags = G_0098F8_MAX_COMPRESSED_FRAGS(gpu_info->gb_addr_config);
-         int log2_samples_frag_diff = MAX2(0, log2_samples - max_compressed_frags);
-         int samples_overlap = MIN2(log2_samples, overlap);
-
-         /* TODO: It shouldn't be necessary if the image has DCC but
-          * not readable by shader.
+         /* Be conservative when the range is unknown because it's not possible to know which mips
+          * are used.
           */
-         if ((radv_image_has_dcc(image) || radv_image_is_tc_compat_cmask(image)) &&
-             (samples_overlap > log2_samples_frag_diff)) {
+         if (first_mip_pipe_misaligned != UINT32_MAX)
             return true;
-         }
       }
    }
 
    return false;
 }
 
-static bool
-radv_image_is_l2_coherent(const struct radv_device *device, const struct radv_image *image)
+bool
+radv_image_is_l2_coherent(const struct radv_device *device, const struct radv_image *image,
+                          const VkImageSubresourceRange *range)
 {
    const struct radv_physical_device *pdev = radv_device_physical(device);
 
    if (pdev->info.gfx_level >= GFX12) {
       return true; /* Everything is coherent with TC L2. */
    } else if (pdev->info.gfx_level >= GFX10) {
-      /* Add a special case for mips in the metadata mip-tail for GFX11. */
-      if (pdev->info.gfx_level >= GFX11) {
-         if (image->vk.mip_levels > 1 && (radv_image_has_dcc(image) || radv_image_has_htile(image))) {
-            for (unsigned i = 0; i < image->plane_count; ++i) {
-               const struct radeon_surf *surf = &image->planes[i].surface;
-
-               if (surf->num_meta_levels != image->vk.mip_levels)
-                  return false;
-            }
-         }
-      }
-
-      return !pdev->info.tcc_rb_non_coherent && !radv_image_is_pipe_misaligned(device, image);
+      return !radv_image_is_pipe_misaligned(image, range);
    } else if (pdev->info.gfx_level == GFX9) {
       if (image->vk.samples == 1 &&
           (image->vk.usage & (VK_IMAGE_USAGE_COLOR_ATTACHMENT_BIT | VK_IMAGE_USAGE_DEPTH_STENCIL_ATTACHMENT_BIT)) &&
@@ -1107,6 +1075,70 @@ radv_surface_init(struct radv_physical_device *pdev, const struct ac_surf_info *
    ac_compute_surface(pdev->addrlib, &pdev->info, &config, mode, surf);
 }
 
+/* Return the first mip level which is pipe-misaligned with metadata, UINT32_MAX means no mips are
+ * affected and zero means all mips.
+ */
+static uint32_t
+radv_image_get_first_mip_pipe_misaligned(const struct radv_device *device, const struct radv_image *image,
+                                         uint32_t plane_id)
+{
+   const struct radv_physical_device *pdev = radv_device_physical(device);
+   const int log2_samples = util_logbase2(image->vk.samples);
+   uint32_t first_mip = UINT32_MAX;
+
+   /* Add a special case for mips in the metadata mip-tail for GFX11. */
+   if (pdev->info.gfx_level >= GFX11) {
+      if (image->vk.mip_levels > 1 && (radv_image_has_dcc(image) || radv_image_has_htile(image))) {
+         first_mip = image->planes[plane_id].surface.num_meta_levels;
+      }
+   }
+
+   VkFormat fmt = radv_image_get_plane_format(pdev, image, plane_id);
+   int log2_bpp = util_logbase2(vk_format_get_blocksize(fmt));
+   int log2_bpp_and_samples;
+
+   if (pdev->info.gfx_level >= GFX10_3) {
+      log2_bpp_and_samples = log2_bpp + log2_samples;
+   } else {
+      if (vk_format_has_depth(image->vk.format) && image->vk.array_layers >= 8) {
+         log2_bpp = 2;
+      }
+
+      log2_bpp_and_samples = MIN2(6, log2_bpp + log2_samples);
+   }
+
+   int num_pipes = G_0098F8_NUM_PIPES(pdev->info.gb_addr_config);
+   int overlap = MAX2(0, log2_bpp_and_samples + num_pipes - 8);
+
+   if (vk_format_has_depth(image->vk.format)) {
+      if (radv_image_is_tc_compat_htile(image) && (pdev->info.tcc_rb_non_coherent || overlap)) {
+         first_mip = 0;
+      }
+   } else {
+      int max_compressed_frags = G_0098F8_MAX_COMPRESSED_FRAGS(pdev->info.gb_addr_config);
+      int log2_samples_frag_diff = MAX2(0, log2_samples - max_compressed_frags);
+      int samples_overlap = MIN2(log2_samples, overlap);
+
+      /* TODO: It shouldn't be necessary if the image has DCC but
+       * not readable by shader.
+       */
+      if ((radv_image_has_dcc(image) || radv_image_is_tc_compat_cmask(image)) &&
+          (pdev->info.tcc_rb_non_coherent || (samples_overlap > log2_samples_frag_diff))) {
+         first_mip = 0;
+      }
+   }
+
+   return first_mip;
+}
+
+static void
+radv_image_init_first_mip_pipe_misaligned(const struct radv_device *device, struct radv_image *image)
+{
+   for (uint32_t i = 0; i < image->plane_count; i++) {
+      image->planes[i].first_mip_pipe_misaligned = radv_image_get_first_mip_pipe_misaligned(device, image, i);
+   }
+}
+
 VkResult
 radv_image_create_layout(struct radv_device *device, struct radv_image_create_info create_info,
                          const struct VkImageDrmFormatModifierExplicitCreateInfoEXT *mod_info,
@@ -1168,6 +1200,16 @@ radv_image_create_layout(struct radv_device *device, struct radv_image_create_in
             ac_surface_zero_dcc_fields(&image->planes[0].surface);
       }
 
+      if (pdev->info.gfx_level >= GFX12 &&
+          (!radv_surface_has_scanout(device, &create_info) || pdev->info.gfx12_supports_display_dcc)) {
+         const enum pipe_format format = vk_format_to_pipe_format(image->vk.format);
+
+         /* Set DCC tilings for both color and depth/stencil. */
+         image->planes[plane].surface.u.gfx9.color.dcc_number_type = ac_get_cb_number_type(format);
+         image->planes[plane].surface.u.gfx9.color.dcc_data_format = ac_get_cb_format(pdev->info.gfx_level, format);
+         image->planes[plane].surface.u.gfx9.color.dcc_write_compress_disable = false;
+      }
+
       if (create_info.bo_metadata && !mod_info &&
           !ac_surface_apply_umd_metadata(&pdev->info, &image->planes[plane].surface, image->vk.samples,
                                          image->vk.mip_levels, create_info.bo_metadata->size_metadata,
@@ -1214,7 +1256,8 @@ radv_image_create_layout(struct radv_device *device, struct radv_image_create_in
 
    image->tc_compatible_cmask = radv_image_has_cmask(image) && radv_use_tc_compat_cmask_for_image(device, image);
 
-   image->l2_coherent = radv_image_is_l2_coherent(device, image);
+   if (pdev->info.gfx_level >= GFX10 && pdev->info.gfx_level < GFX12)
+      radv_image_init_first_mip_pipe_misaligned(device, image);
 
    image->support_comp_to_single = radv_image_use_comp_to_single(device, image);
 
@@ -1222,7 +1265,9 @@ radv_image_create_layout(struct radv_device *device, struct radv_image_create_in
 
    assert(image->planes[0].surface.surf_size);
    assert(image->planes[0].surface.modifier == DRM_FORMAT_MOD_INVALID ||
-          ac_modifier_has_dcc(image->planes[0].surface.modifier) == radv_image_has_dcc(image));
+          ac_modifier_has_dcc(image->planes[0].surface.modifier) ==
+             (pdev->info.gfx_level >= GFX12 ? image->planes[0].surface.u.gfx9.gfx12_enable_dcc
+                                            : radv_image_has_dcc(image)));
    return VK_SUCCESS;
 }
 
@@ -1714,7 +1759,7 @@ radv_BindImageMemory2(VkDevice _device, uint32_t bindInfoCount, const VkBindImag
    for (uint32_t i = 0; i < bindInfoCount; ++i) {
       VK_FROM_HANDLE(radv_device_memory, mem, pBindInfos[i].memory);
       VK_FROM_HANDLE(radv_image, image, pBindInfos[i].image);
-      VkBindMemoryStatusKHR *status = (void *)vk_find_struct_const(&pBindInfos[i], BIND_MEMORY_STATUS_KHR);
+      VkBindMemoryStatus *status = (void *)vk_find_struct_const(&pBindInfos[i], BIND_MEMORY_STATUS);
 
       if (status)
          *status->pResult = VK_SUCCESS;
@@ -1773,8 +1818,8 @@ radv_BindImageMemory2(VkDevice _device, uint32_t bindInfoCount, const VkBindImag
 }
 
 VKAPI_ATTR void VKAPI_CALL
-radv_GetImageSubresourceLayout2KHR(VkDevice _device, VkImage _image, const VkImageSubresource2KHR *pSubresource,
-                                   VkSubresourceLayout2KHR *pLayout)
+radv_GetImageSubresourceLayout2(VkDevice _device, VkImage _image, const VkImageSubresource2 *pSubresource,
+                                VkSubresourceLayout2 *pLayout)
 {
    VK_FROM_HANDLE(radv_image, image, _image);
    VK_FROM_HANDLE(radv_device, device, _device);
