@@ -29,15 +29,16 @@
 #include <xcb/dri3.h>
 #include <xcb/present.h>
 #include <xcb/xcb.h>
-#include <xcb/xfixes.h>
 
 #include <xf86drm.h>
+#include "drm-uapi/drm_fourcc.h"
 #include "util/macros.h"
 
 #include "egl_dri2.h"
 #include "platform_x11_dri3.h"
 
 #include "loader.h"
+#include "loader_x11.h"
 #include "loader_dri3_helper.h"
 
 static struct dri3_egl_surface *
@@ -66,7 +67,7 @@ egl_dri3_in_current_context(struct loader_dri3_drawable *draw)
    return ctx->Resource.Display == dri3_surf->surf.base.Resource.Display;
 }
 
-static __DRIcontext *
+static struct dri_context *
 egl_dri3_get_dri_context(struct loader_dri3_drawable *draw)
 {
    _EGLContext *ctx = _eglGetCurrentContext();
@@ -77,7 +78,7 @@ egl_dri3_get_dri_context(struct loader_dri3_drawable *draw)
    return dri2_ctx->dri_context;
 }
 
-static __DRIscreen *
+static struct dri_screen *
 egl_dri3_get_dri_screen(void)
 {
    _EGLContext *ctx = _eglGetCurrentContext();
@@ -157,7 +158,7 @@ dri3_create_surface(_EGLDisplay *disp, EGLint type, _EGLConfig *conf,
    struct dri2_egl_display *dri2_dpy = dri2_egl_display(disp);
    struct dri2_egl_config *dri2_conf = dri2_egl_config(conf);
    struct dri3_egl_surface *dri3_surf;
-   const __DRIconfig *dri_config;
+   const struct dri_config *dri_config;
    xcb_drawable_t drawable;
 
    dri3_surf = calloc(1, sizeof *dri3_surf);
@@ -193,7 +194,7 @@ dri3_create_surface(_EGLDisplay *disp, EGLint type, _EGLConfig *conf,
           dri2_dpy->conn, drawable, egl_to_loader_dri3_drawable_type(type),
           dri2_dpy->dri_screen_render_gpu, dri2_dpy->dri_screen_display_gpu,
           dri2_dpy->multibuffers_available, true, dri_config,
-          &dri2_dpy->loader_dri3_ext, &egl_dri3_vtable,
+          &egl_dri3_vtable,
           &dri3_surf->loader_drawable)) {
       _eglError(EGL_BAD_ALLOC, "dri3_surface_create");
       goto cleanup_pixmap;
@@ -225,7 +226,7 @@ dri3_authenticate(_EGLDisplay *disp, uint32_t id)
 #ifdef HAVE_WAYLAND_PLATFORM
    struct dri2_egl_display *dri2_dpy = dri2_egl_display(disp);
 
-   if (dri2_dpy->device_name) {
+   if (!dri2_dpy->swrast) {
       _eglLog(_EGL_WARNING,
               "Wayland client render node authentication is unnecessary");
       return 0;
@@ -293,7 +294,7 @@ dri3_create_image_khr_pixmap(_EGLDisplay *disp, _EGLContext *ctx,
    xcb_drawable_t drawable;
    xcb_dri3_buffer_from_pixmap_cookie_t bp_cookie;
    xcb_dri3_buffer_from_pixmap_reply_t *bp_reply;
-   unsigned int format;
+   unsigned int fourcc;
 
    drawable = (xcb_drawable_t)(uintptr_t)buffer;
    bp_cookie = xcb_dri3_buffer_from_pixmap(dri2_dpy->conn, drawable);
@@ -304,8 +305,8 @@ dri3_create_image_khr_pixmap(_EGLDisplay *disp, _EGLContext *ctx,
       return NULL;
    }
 
-   format = dri2_format_for_depth(dri2_dpy, bp_reply->depth);
-   if (format == __DRI_IMAGE_FORMAT_NONE) {
+   fourcc = dri2_fourcc_for_depth(dri2_dpy, bp_reply->depth);
+   if (fourcc == DRM_FORMAT_INVALID) {
       _eglError(EGL_BAD_PARAMETER,
                 "dri3_create_image_khr: unsupported pixmap depth");
       free(bp_reply);
@@ -322,15 +323,15 @@ dri3_create_image_khr_pixmap(_EGLDisplay *disp, _EGLContext *ctx,
    _eglInitImage(&dri2_img->base, disp);
 
    dri2_img->dri_image = loader_dri3_create_image(
-      dri2_dpy->conn, bp_reply, format, dri2_dpy->dri_screen_render_gpu,
-      dri2_dpy->image, dri2_img);
+      dri2_dpy->conn, bp_reply, fourcc, dri2_dpy->dri_screen_render_gpu,
+      dri2_img);
 
    free(bp_reply);
 
    return &dri2_img->base;
 }
 
-#ifdef HAVE_DRI3_MODIFIERS
+#ifdef HAVE_X11_DRM
 static _EGLImage *
 dri3_create_image_khr_pixmap_from_buffers(_EGLDisplay *disp, _EGLContext *ctx,
                                           EGLClientBuffer buffer,
@@ -341,7 +342,7 @@ dri3_create_image_khr_pixmap_from_buffers(_EGLDisplay *disp, _EGLContext *ctx,
    xcb_dri3_buffers_from_pixmap_cookie_t bp_cookie;
    xcb_dri3_buffers_from_pixmap_reply_t *bp_reply;
    xcb_drawable_t drawable;
-   unsigned int format;
+   unsigned int fourcc;
 
    drawable = (xcb_drawable_t)(uintptr_t)buffer;
    bp_cookie = xcb_dri3_buffers_from_pixmap(dri2_dpy->conn, drawable);
@@ -353,8 +354,8 @@ dri3_create_image_khr_pixmap_from_buffers(_EGLDisplay *disp, _EGLContext *ctx,
       return EGL_NO_IMAGE_KHR;
    }
 
-   format = dri2_format_for_depth(dri2_dpy, bp_reply->depth);
-   if (format == __DRI_IMAGE_FORMAT_NONE) {
+   fourcc = dri2_fourcc_for_depth(dri2_dpy, bp_reply->depth);
+   if (fourcc == DRM_FORMAT_INVALID) {
       _eglError(EGL_BAD_PARAMETER,
                 "dri3_create_image_khr: unsupported pixmap depth");
       free(bp_reply);
@@ -371,8 +372,8 @@ dri3_create_image_khr_pixmap_from_buffers(_EGLDisplay *disp, _EGLContext *ctx,
    _eglInitImage(&dri2_img->base, disp);
 
    dri2_img->dri_image = loader_dri3_create_image_from_buffers(
-      dri2_dpy->conn, bp_reply, format, dri2_dpy->dri_screen_render_gpu,
-      dri2_dpy->image, dri2_img);
+      dri2_dpy->conn, bp_reply, fourcc, dri2_dpy->dri_screen_render_gpu,
+      dri2_img);
    free(bp_reply);
 
    if (!dri2_img->dri_image) {
@@ -389,13 +390,13 @@ static _EGLImage *
 dri3_create_image_khr(_EGLDisplay *disp, _EGLContext *ctx, EGLenum target,
                       EGLClientBuffer buffer, const EGLint *attr_list)
 {
-#ifdef HAVE_DRI3_MODIFIERS
+#ifdef HAVE_X11_DRM
    struct dri2_egl_display *dri2_dpy = dri2_egl_display(disp);
 #endif
 
    switch (target) {
    case EGL_NATIVE_PIXMAP_KHR:
-#ifdef HAVE_DRI3_MODIFIERS
+#ifdef HAVE_X11_DRM
       if (dri2_dpy->multibuffers_available)
          return dri3_create_image_khr_pixmap_from_buffers(disp, ctx, buffer,
                                                           attr_list);
@@ -411,7 +412,7 @@ dri3_create_image_khr(_EGLDisplay *disp, _EGLContext *ctx, EGLenum target,
  * contents of its fake front buffer.
  */
 static void
-dri3_flush_front_buffer(__DRIdrawable *driDrawable, void *loaderPrivate)
+dri3_flush_front_buffer(struct dri_drawable *driDrawable, void *loaderPrivate)
 {
    struct loader_dri3_drawable *draw = loaderPrivate;
    (void)driDrawable;
@@ -491,7 +492,7 @@ dri3_query_surface(_EGLDisplay *disp, _EGLSurface *surf, EGLint attribute,
    return _eglQuerySurface(disp, surf, attribute, value);
 }
 
-static __DRIdrawable *
+static struct dri_drawable *
 dri3_get_dri_drawable(_EGLSurface *surf)
 {
    struct dri3_egl_surface *dri3_surf = dri3_egl_surface(surf);
@@ -526,105 +527,19 @@ struct dri2_egl_display_vtbl dri3_x11_display_vtbl = {
    .close_screen_notify = dri3_close_screen_notify,
 };
 
-/* Only request versions of these protocols which we actually support. */
-#define DRI3_SUPPORTED_MAJOR    1
-#define PRESENT_SUPPORTED_MAJOR 1
-
-#ifdef HAVE_DRI3_MODIFIERS
-#define DRI3_SUPPORTED_MINOR    2
-#define PRESENT_SUPPORTED_MINOR 2
-#else
-#define PRESENT_SUPPORTED_MINOR 0
-#define DRI3_SUPPORTED_MINOR    0
-#endif
-
 enum dri2_egl_driver_fail
-dri3_x11_connect(struct dri2_egl_display *dri2_dpy, bool zink)
+dri3_x11_connect(struct dri2_egl_display *dri2_dpy, bool zink, bool swrast)
 {
-   xcb_dri3_query_version_reply_t *dri3_query;
-   xcb_dri3_query_version_cookie_t dri3_query_cookie;
-   xcb_present_query_version_reply_t *present_query;
-   xcb_present_query_version_cookie_t present_query_cookie;
-   xcb_xfixes_query_version_reply_t *xfixes_query;
-   xcb_xfixes_query_version_cookie_t xfixes_query_cookie;
-   xcb_generic_error_t *error;
-   const xcb_query_extension_reply_t *extension;
-
-   dri2_dpy->dri3_major_version = 0;
-   dri2_dpy->dri3_minor_version = 0;
-   dri2_dpy->present_major_version = 0;
-   dri2_dpy->present_minor_version = 0;
-
-   xcb_prefetch_extension_data(dri2_dpy->conn, &xcb_dri3_id);
-   xcb_prefetch_extension_data(dri2_dpy->conn, &xcb_present_id);
-   xcb_prefetch_extension_data(dri2_dpy->conn, &xcb_xfixes_id);
-
-   extension = xcb_get_extension_data(dri2_dpy->conn, &xcb_dri3_id);
-   if (!(extension && extension->present))
-      return DRI2_EGL_DRIVER_FAILED;
-
-   extension = xcb_get_extension_data(dri2_dpy->conn, &xcb_present_id);
-   if (!(extension && extension->present))
-      return DRI2_EGL_DRIVER_FAILED;
-
-   extension = xcb_get_extension_data(dri2_dpy->conn, &xcb_xfixes_id);
-   if (!(extension && extension->present))
-      return DRI2_EGL_DRIVER_FAILED;
-
-   dri3_query_cookie = xcb_dri3_query_version(
-      dri2_dpy->conn, DRI3_SUPPORTED_MAJOR, DRI3_SUPPORTED_MINOR);
-
-   present_query_cookie = xcb_present_query_version(
-      dri2_dpy->conn, PRESENT_SUPPORTED_MAJOR, PRESENT_SUPPORTED_MINOR);
-
-   xfixes_query_cookie = xcb_xfixes_query_version(
-      dri2_dpy->conn, XCB_XFIXES_MAJOR_VERSION, XCB_XFIXES_MINOR_VERSION);
-
-   dri3_query =
-      xcb_dri3_query_version_reply(dri2_dpy->conn, dri3_query_cookie, &error);
-   if (dri3_query == NULL || error != NULL) {
-      _eglLog(_EGL_WARNING, "DRI3: failed to query the version");
-      free(dri3_query);
-      free(error);
-      return DRI2_EGL_DRIVER_FAILED;
-   }
-
-   dri2_dpy->dri3_major_version = dri3_query->major_version;
-   dri2_dpy->dri3_minor_version = dri3_query->minor_version;
-   free(dri3_query);
-
-   present_query = xcb_present_query_version_reply(
-      dri2_dpy->conn, present_query_cookie, &error);
-   if (present_query == NULL || error != NULL) {
-      _eglLog(_EGL_WARNING, "DRI3: failed to query Present version");
-      free(present_query);
-      free(error);
-      return DRI2_EGL_DRIVER_FAILED;
-   }
-
-   dri2_dpy->present_major_version = present_query->major_version;
-   dri2_dpy->present_minor_version = present_query->minor_version;
-   free(present_query);
-
-   xfixes_query = xcb_xfixes_query_version_reply(dri2_dpy->conn,
-                                                 xfixes_query_cookie, &error);
-   if (xfixes_query == NULL || error != NULL ||
-       xfixes_query->major_version < 2) {
-      _eglLog(_EGL_WARNING, "DRI3: failed to query xfixes version");
-      free(error);
-      free(xfixes_query);
-      return DRI2_EGL_DRIVER_FAILED;
-   }
-   free(xfixes_query);
-
    dri2_dpy->fd_render_gpu =
-      loader_dri3_open(dri2_dpy->conn, dri2_dpy->screen->root, 0);
+      x11_dri3_open(dri2_dpy->conn, dri2_dpy->screen->root, 0);
    if (dri2_dpy->fd_render_gpu < 0) {
       int conn_error = xcb_connection_has_error(dri2_dpy->conn);
-      _eglLog(_EGL_WARNING, "DRI3: Screen seems not DRI3 capable");
+      if (!swrast) {
+         _eglLog(_EGL_INFO, "DRI3: Could not get DRI3 device");
 
-      if (conn_error)
-         _eglLog(_EGL_WARNING, "DRI3: Failed to initialize");
+         if (conn_error)
+            _eglLog(_EGL_WARNING, "DRI3: Failed to initialize");
+      }
 
       return DRI2_EGL_DRIVER_FAILED;
    }
