@@ -73,8 +73,9 @@ enum tu_cmd_dirty_bits
    TU_CMD_DIRTY_RAST_ORDER = BIT(12),
    TU_CMD_DIRTY_FEEDBACK_LOOPS = BIT(13),
    TU_CMD_DIRTY_FS = BIT(14),
+   TU_CMD_DIRTY_SHADING_RATE = BIT(15),
    /* all draw states were disabled and need to be re-enabled: */
-   TU_CMD_DIRTY_DRAW_STATE = BIT(15)
+   TU_CMD_DIRTY_DRAW_STATE = BIT(16)
 };
 
 /* There are only three cache domains we have to care about: the CCU, or
@@ -147,6 +148,8 @@ enum tu_cmd_access_mask {
     */
    TU_ACCESS_CCHE_READ = 1 << 16,
 
+   TU_ACCESS_RTU_READ = 1 << 17,
+
    TU_ACCESS_READ =
       TU_ACCESS_UCHE_READ |
       TU_ACCESS_CCU_COLOR_READ |
@@ -211,6 +214,7 @@ enum tu_cmd_flush_bits {
     * as it isn't necessary. Therefore, it's not included in ALL_FLUSH.
     */
    TU_CMD_FLAG_BLIT_CACHE_CLEAN = 1 << 11,
+   TU_CMD_FLAG_RTU_INVALIDATE = 1 << 12,
 
    TU_CMD_FLAG_ALL_CLEAN =
       TU_CMD_FLAG_CCU_CLEAN_DEPTH |
@@ -233,7 +237,8 @@ enum tu_cmd_flush_bits {
        * in case there was another command before the current command buffer
        * that it needs to wait for.
        */
-      TU_CMD_FLAG_WAIT_FOR_ME,
+      TU_CMD_FLAG_WAIT_FOR_ME |
+      TU_CMD_FLAG_RTU_INVALIDATE,
 };
 
 /* Changing the CCU from sysmem mode to gmem mode or vice-versa is pretty
@@ -314,6 +319,7 @@ struct tu_render_pass_state
    uint32_t drawcall_bandwidth_per_sample_sum;
 
    const char *lrz_disable_reason;
+   uint32_t lrz_disabled_at_draw;
 };
 
 /* These are the states of the suspend/resume state machine. In addition to
@@ -514,6 +520,9 @@ struct tu_cmd_state
    bool raster_order_attachment_access_valid;
    bool blit_cache_cleaned;
    VkImageAspectFlags pipeline_feedback_loops;
+   bool pipeline_writes_shading_rate;
+   bool pipeline_reads_shading_rate;
+   bool pipeline_accesses_smask;
 
    bool pipeline_blend_lrz, pipeline_bandwidth;
    uint32_t pipeline_draw_states;
@@ -571,11 +580,11 @@ struct tu_cmd_buffer
 
    struct tu_descriptor_state descriptors[MAX_BIND_POINTS];
 
-   struct tu_render_pass_attachment dynamic_rp_attachments[2 * (MAX_RTS + 1) + 1];
+   struct tu_render_pass_attachment dynamic_rp_attachments[2 * (MAX_RTS + 1) + 2];
    struct tu_subpass_attachment dynamic_color_attachments[MAX_RTS];
    struct tu_subpass_attachment dynamic_input_attachments[MAX_RTS + 1];
    struct tu_subpass_attachment dynamic_resolve_attachments[MAX_RTS + 1];
-   const struct tu_image_view *dynamic_attachments[2 * (MAX_RTS + 1) + 1];
+   const struct tu_image_view *dynamic_attachments[2 * (MAX_RTS + 1) + 2];
    VkClearValue dynamic_clear_values[2 * (MAX_RTS + 1)];
 
    struct tu_render_pass dynamic_pass;
@@ -613,6 +622,8 @@ struct tu_cmd_buffer
    uint32_t vsc_prim_strm_pitch;
    uint64_t vsc_draw_strm_va, vsc_draw_strm_size_va, vsc_prim_strm_va;
    bool vsc_initialized;
+
+   bool prev_fsr_is_null;
 };
 VK_DEFINE_HANDLE_CASTS(tu_cmd_buffer, vk.base, VkCommandBuffer,
                        VK_OBJECT_TYPE_COMMAND_BUFFER)
@@ -677,8 +688,27 @@ tu_restore_suspended_pass(struct tu_cmd_buffer *cmd,
 template <chip CHIP>
 void tu_cmd_render(struct tu_cmd_buffer *cmd);
 
+void tu_dispatch_unaligned(VkCommandBuffer commandBuffer,
+                           uint32_t x, uint32_t y, uint32_t z);
+
+void tu_dispatch_unaligned_indirect(VkCommandBuffer commandBuffer,
+                                    VkDeviceAddress size_addr);
+
+void tu_write_buffer_cp(VkCommandBuffer commandBuffer,
+                        VkDeviceAddress addr,
+                        void *data, uint32_t size);
+
+void tu_flush_buffer_write_cp(VkCommandBuffer commandBuffer);
+
 enum fd_gpu_event : uint32_t;
 
+template <chip CHIP>
+void
+tu_emit_raw_event_write(struct tu_cmd_buffer *cmd,
+                        struct tu_cs *cs,
+                        enum vgt_event_type event,
+                        bool needs_seqno);
+
 template <chip CHIP>
 void
 tu_emit_event_write(struct tu_cmd_buffer *cmd,
