@@ -22,8 +22,7 @@
  * IN THE SOFTWARE.
  */
 
-#ifndef BRW_IR_FS_H
-#define BRW_IR_FS_H
+#pragma once
 
 #include "brw_ir.h"
 #include "brw_ir_allocator.h"
@@ -58,7 +57,7 @@ public:
    bool is_payload(unsigned arg) const;
    bool is_partial_write() const;
    unsigned components_read(unsigned i) const;
-   unsigned size_read(int arg) const;
+   unsigned size_read(const struct intel_device_info *devinfo, int arg) const;
    bool can_do_source_mods(const struct intel_device_info *devinfo) const;
    bool can_do_cmod() const;
    bool can_change_types() const;
@@ -123,6 +122,12 @@ public:
     */
    bool has_sampler_residency() const;
 
+   /**
+    * Return true if this instruction is using the address register
+    * implicitly.
+    */
+   bool uses_address_register_implicitly() const;
+
    uint8_t sources; /**< Number of brw_reg sources. */
 
    /**
@@ -188,10 +193,6 @@ public:
          bool check_tdr:1; /**< Only valid for SEND; turns it into a SENDC */
          bool send_has_side_effects:1; /**< Only valid for SHADER_OPCODE_SEND */
          bool send_is_volatile:1; /**< Only valid for SHADER_OPCODE_SEND */
-         bool send_ex_desc_scratch:1; /**< Only valid for SHADER_OPCODE_SEND, use
-                                       *   the scratch surface offset to build
-                                       *   extended descriptor
-                                       */
          bool send_ex_bso:1; /**< Only for SHADER_OPCODE_SEND, use extended
                               *   bindless surface offset (26bits instead of
                               *   20bits)
@@ -296,15 +297,15 @@ regs_written(const fs_inst *inst)
  * UNIFORM files and 32B for all other files.
  */
 inline unsigned
-regs_read(const fs_inst *inst, unsigned i)
+regs_read(const struct intel_device_info *devinfo, const fs_inst *inst, unsigned i)
 {
    if (inst->src[i].file == IMM)
       return 1;
 
    const unsigned reg_size = inst->src[i].file == UNIFORM ? 4 : REG_SIZE;
    return DIV_ROUND_UP(reg_offset(inst->src[i]) % reg_size +
-                       inst->size_read(i) -
-                       MIN2(inst->size_read(i), reg_padding(inst->src[i])),
+                       inst->size_read(devinfo, i) -
+                       MIN2(inst->size_read(devinfo, i), reg_padding(inst->src[i])),
                        reg_size);
 }
 
@@ -476,7 +477,8 @@ has_subdword_integer_region_restriction(const intel_device_info *devinfo,
  * multiple virtual registers in any order is allowed.
  */
 inline bool
-is_copy_payload(brw_reg_file file, const fs_inst *inst)
+is_copy_payload(const struct intel_device_info *devinfo,
+                brw_reg_file file, const fs_inst *inst)
 {
    if (inst->opcode != SHADER_OPCODE_LOAD_PAYLOAD ||
        inst->is_partial_write() || inst->saturate ||
@@ -492,7 +494,7 @@ is_copy_payload(brw_reg_file file, const fs_inst *inst)
          return false;
 
       if (regions_overlap(inst->dst, inst->size_written,
-                          inst->src[i], inst->size_read(i)))
+                          inst->src[i], inst->size_read(devinfo, i)))
          return false;
    }
 
@@ -505,8 +507,10 @@ is_copy_payload(brw_reg_file file, const fs_inst *inst)
  * destination without any reordering.
  */
 inline bool
-is_identity_payload(brw_reg_file file, const fs_inst *inst) {
-   if (is_copy_payload(file, inst)) {
+is_identity_payload(const struct intel_device_info *devinfo,
+                    brw_reg_file file, const fs_inst *inst)
+{
+   if (is_copy_payload(devinfo, file, inst)) {
       brw_reg reg = inst->src[0];
 
       for (unsigned i = 0; i < inst->sources; i++) {
@@ -514,7 +518,7 @@ is_identity_payload(brw_reg_file file, const fs_inst *inst) {
          if (!inst->src[i].equals(reg))
             return false;
 
-         reg = byte_offset(reg, inst->size_read(i));
+         reg = byte_offset(reg, inst->size_read(devinfo, i));
       }
 
       return true;
@@ -534,8 +538,10 @@ is_identity_payload(brw_reg_file file, const fs_inst *inst) {
  * instructions.
  */
 inline bool
-is_multi_copy_payload(const fs_inst *inst) {
-   if (is_copy_payload(VGRF, inst)) {
+is_multi_copy_payload(const struct intel_device_info *devinfo,
+                      const fs_inst *inst)
+{
+   if (is_copy_payload(devinfo, VGRF, inst)) {
       for (unsigned i = 0; i < inst->sources; i++) {
             if (inst->src[i].nr != inst->src[0].nr)
                return true;
@@ -558,9 +564,10 @@ is_multi_copy_payload(const fs_inst *inst) {
  * instruction.
  */
 inline bool
-is_coalescing_payload(const brw::simple_allocator &alloc, const fs_inst *inst)
+is_coalescing_payload(const struct intel_device_info *devinfo,
+                      const brw::simple_allocator &alloc, const fs_inst *inst)
 {
-   return is_identity_payload(VGRF, inst) &&
+   return is_identity_payload(devinfo, VGRF, inst) &&
           inst->src[0].offset == 0 &&
           alloc.sizes[inst->src[0].nr] * REG_SIZE == inst->size_written;
 }
@@ -599,5 +606,3 @@ brw_fs_flag_mask(const brw_reg &r, unsigned sz)
       return 0;
    }
 }
-
-#endif
